import pytest
from flexmock import flexmock

from borgmatic.commands import arguments as module


def test_parse_arguments_with_no_arguments_uses_defaults():
    config_paths = ['default']
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(config_paths)

    arguments = module.parse_arguments()

    global_arguments = arguments['global']
    assert global_arguments.config_paths == config_paths
    assert global_arguments.verbosity == 0
    assert global_arguments.syslog_verbosity == -2
    assert global_arguments.log_file_verbosity == 1
    assert global_arguments.monitoring_verbosity == 1


def test_parse_arguments_with_multiple_config_flags_parses_as_list():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('--config', 'myconfig', '--config', 'otherconfig')

    global_arguments = arguments['global']
    assert global_arguments.config_paths == ['myconfig', 'otherconfig']
    assert global_arguments.verbosity == 0
    assert global_arguments.syslog_verbosity == -2
    assert global_arguments.log_file_verbosity == 1
    assert global_arguments.monitoring_verbosity == 1


def test_parse_arguments_with_action_after_config_path_omits_action():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('--config', 'myconfig', 'list', '--json')

    global_arguments = arguments['global']
    assert global_arguments.config_paths == ['myconfig']
    assert 'list' in arguments
    assert arguments['list'].json


def test_parse_arguments_with_action_after_config_path_omits_aliased_action():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('--config', 'myconfig', 'init', '--encryption', 'repokey')

    global_arguments = arguments['global']
    assert global_arguments.config_paths == ['myconfig']
    assert 'rcreate' in arguments
    assert arguments['rcreate'].encryption_mode == 'repokey'


def test_parse_arguments_with_action_and_positional_arguments_after_config_path_omits_action_and_arguments():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('--config', 'myconfig', 'borg', 'key', 'export')

    global_arguments = arguments['global']
    assert global_arguments.config_paths == ['myconfig']
    assert 'borg' in arguments
    assert arguments['borg'].options == ['key', 'export']


def test_parse_arguments_with_verbosity_overrides_default():
    config_paths = ['default']
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(config_paths)

    arguments = module.parse_arguments('--verbosity', '1')

    global_arguments = arguments['global']
    assert global_arguments.config_paths == config_paths
    assert global_arguments.verbosity == 1
    assert global_arguments.syslog_verbosity == -2
    assert global_arguments.log_file_verbosity == 1
    assert global_arguments.monitoring_verbosity == 1


def test_parse_arguments_with_syslog_verbosity_overrides_default():
    config_paths = ['default']
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(config_paths)

    arguments = module.parse_arguments('--syslog-verbosity', '2')

    global_arguments = arguments['global']
    assert global_arguments.config_paths == config_paths
    assert global_arguments.verbosity == 0
    assert global_arguments.syslog_verbosity == 2
    assert global_arguments.log_file_verbosity == 1
    assert global_arguments.monitoring_verbosity == 1


def test_parse_arguments_with_log_file_verbosity_overrides_default():
    config_paths = ['default']
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(config_paths)

    arguments = module.parse_arguments('--log-file-verbosity', '-1')

    global_arguments = arguments['global']
    assert global_arguments.config_paths == config_paths
    assert global_arguments.verbosity == 0
    assert global_arguments.syslog_verbosity == -2
    assert global_arguments.log_file_verbosity == -1
    assert global_arguments.monitoring_verbosity == 1


def test_parse_arguments_with_single_override_parses():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('--override', 'foo.bar=baz')

    global_arguments = arguments['global']
    assert global_arguments.overrides == ['foo.bar=baz']


def test_parse_arguments_with_multiple_overrides_flags_parses():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments(
        '--override', 'foo.bar=baz', '--override', 'foo.quux=7', '--override', 'this.that=8'
    )

    global_arguments = arguments['global']
    assert global_arguments.overrides == ['foo.bar=baz', 'foo.quux=7', 'this.that=8']


def test_parse_arguments_with_list_json_overrides_default():
    arguments = module.parse_arguments('list', '--json')

    assert 'list' in arguments
    assert arguments['list'].json is True


def test_parse_arguments_with_no_actions_defaults_to_all_actions_enabled():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments()

    assert 'prune' in arguments
    assert 'create' in arguments
    assert 'check' in arguments


def test_parse_arguments_with_no_actions_passes_argument_to_relevant_actions():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('--stats', '--list')

    assert 'prune' in arguments
    assert arguments['prune'].stats
    assert arguments['prune'].list_archives
    assert 'create' in arguments
    assert arguments['create'].stats
    assert arguments['create'].list_files
    assert 'check' in arguments


def test_parse_arguments_with_help_and_no_actions_shows_global_help(capsys):
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit) as exit:
        module.parse_arguments('--help')

    assert exit.value.code == 0
    captured = capsys.readouterr()
    assert 'global arguments:' in captured.out
    assert 'actions:' in captured.out


def test_parse_arguments_with_help_and_action_shows_action_help(capsys):
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit) as exit:
        module.parse_arguments('create', '--help')

    assert exit.value.code == 0
    captured = capsys.readouterr()
    assert 'global arguments:' not in captured.out
    assert 'actions:' not in captured.out
    assert 'create arguments:' in captured.out


def test_parse_arguments_with_action_before_global_options_parses_options():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('prune', '--verbosity', '2')

    assert 'prune' in arguments
    assert arguments['global'].verbosity == 2


def test_parse_arguments_with_global_options_before_action_parses_options():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('--verbosity', '2', 'prune')

    assert 'prune' in arguments
    assert arguments['global'].verbosity == 2


def test_parse_arguments_with_prune_action_leaves_other_actions_disabled():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('prune')

    assert 'prune' in arguments
    assert 'create' not in arguments
    assert 'check' not in arguments


def test_parse_arguments_with_multiple_actions_leaves_other_action_disabled():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    arguments = module.parse_arguments('create', 'check')

    assert 'prune' not in arguments
    assert 'create' in arguments
    assert 'check' in arguments


def test_parse_arguments_disallows_invalid_argument():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--posix-me-harder')


def test_parse_arguments_disallows_encryption_mode_without_init():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--config', 'myconfig', '--encryption', 'repokey')


def test_parse_arguments_allows_encryption_mode_with_init():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--config', 'myconfig', 'init', '--encryption', 'repokey')


def test_parse_arguments_requires_encryption_mode_with_init():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit):
        module.parse_arguments('--config', 'myconfig', 'init')


def test_parse_arguments_disallows_append_only_without_init():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--config', 'myconfig', '--append-only')


def test_parse_arguments_disallows_storage_quota_without_init():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--config', 'myconfig', '--storage-quota', '5G')


def test_parse_arguments_allows_init_and_prune():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--config', 'myconfig', 'init', '--encryption', 'repokey', 'prune')


def test_parse_arguments_allows_init_and_create():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--config', 'myconfig', 'init', '--encryption', 'repokey', 'create')


def test_parse_arguments_allows_repository_with_extract():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments(
        '--config', 'myconfig', 'extract', '--repository', 'test.borg', '--archive', 'test'
    )


def test_parse_arguments_allows_repository_with_mount():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments(
        '--config',
        'myconfig',
        'mount',
        '--repository',
        'test.borg',
        '--archive',
        'test',
        '--mount-point',
        '/mnt',
    )


def test_parse_arguments_allows_repository_with_list():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--config', 'myconfig', 'list', '--repository', 'test.borg')


def test_parse_arguments_disallows_archive_unless_action_consumes_it():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--config', 'myconfig', '--archive', 'test')


def test_parse_arguments_disallows_paths_unless_action_consumes_it():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--config', 'myconfig', '--path', 'test')


def test_parse_arguments_disallows_other_actions_with_config_bootstrap():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('config', 'bootstrap', '--repository', 'test.borg', 'list')


def test_parse_arguments_allows_archive_with_extract():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--config', 'myconfig', 'extract', '--archive', 'test')


def test_parse_arguments_allows_archive_with_mount():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments(
        '--config', 'myconfig', 'mount', '--archive', 'test', '--mount-point', '/mnt'
    )


def test_parse_arguments_allows_archive_with_restore():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--config', 'myconfig', 'restore', '--archive', 'test')


def test_parse_arguments_allows_archive_with_list():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--config', 'myconfig', 'list', '--archive', 'test')


def test_parse_arguments_requires_archive_with_extract():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit):
        module.parse_arguments('--config', 'myconfig', 'extract')


def test_parse_arguments_requires_archive_with_restore():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit):
        module.parse_arguments('--config', 'myconfig', 'restore')


def test_parse_arguments_requires_mount_point_with_mount():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit):
        module.parse_arguments('--config', 'myconfig', 'mount', '--archive', 'test')


def test_parse_arguments_requires_mount_point_with_umount():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit):
        module.parse_arguments('--config', 'myconfig', 'umount')


def test_parse_arguments_allows_progress_before_create():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--progress', 'create', 'list')


def test_parse_arguments_allows_progress_after_create():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('create', '--progress', 'list')


def test_parse_arguments_allows_progress_and_extract():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--progress', 'extract', '--archive', 'test', 'list')


def test_parse_arguments_disallows_progress_without_create():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--progress', 'list')


def test_parse_arguments_with_stats_and_create_flags_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--stats', 'create', 'list')


def test_parse_arguments_with_stats_and_prune_flags_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--stats', 'prune', 'list')


def test_parse_arguments_with_stats_flag_but_no_create_or_prune_flag_raises_value_error():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--stats', 'list')


def test_parse_arguments_with_list_and_create_flags_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--list', 'create')


def test_parse_arguments_with_list_and_prune_flags_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--list', 'prune')


def test_parse_arguments_with_list_flag_but_no_relevant_action_raises_value_error():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit):
        module.parse_arguments('--list', 'rcreate')


def test_parse_arguments_disallows_list_with_progress_for_create_action():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('create', '--list', '--progress')


def test_parse_arguments_disallows_list_with_json_for_create_action():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('create', '--list', '--json')


def test_parse_arguments_allows_json_with_list_or_info():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('list', '--json')
    module.parse_arguments('info', '--json')


def test_parse_arguments_disallows_json_with_both_list_and_info():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('list', 'info', '--json')


def test_parse_arguments_disallows_json_with_both_list_and_rinfo():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('list', 'rinfo', '--json')


def test_parse_arguments_disallows_json_with_both_rinfo_and_info():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('rinfo', 'info', '--json')


def test_parse_arguments_disallows_transfer_with_both_archive_and_match_archives():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments(
            'transfer',
            '--source-repository',
            'source.borg',
            '--archive',
            'foo',
            '--match-archives',
            'sh:*bar',
        )


def test_parse_arguments_disallows_list_with_both_prefix_and_match_archives():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('list', '--prefix', 'foo', '--match-archives', 'sh:*bar')


def test_parse_arguments_disallows_rlist_with_both_prefix_and_match_archives():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('rlist', '--prefix', 'foo', '--match-archives', 'sh:*bar')


def test_parse_arguments_disallows_info_with_both_archive_and_match_archives():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('info', '--archive', 'foo', '--match-archives', 'sh:*bar')


def test_parse_arguments_disallows_info_with_both_archive_and_prefix():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('info', '--archive', 'foo', '--prefix', 'bar')


def test_parse_arguments_disallows_info_with_both_prefix_and_match_archives():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('info', '--prefix', 'foo', '--match-archives', 'sh:*bar')


def test_parse_arguments_check_only_extract_does_not_raise_extract_subparser_error():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('check', '--only', 'extract')


def test_parse_arguments_extract_archive_check_does_not_raise_check_subparser_error():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('extract', '--archive', 'check')


def test_parse_arguments_extract_with_check_only_extract_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('extract', '--archive', 'name', 'check', '--only', 'extract')


def test_parse_arguments_bootstrap_without_config_errors():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('bootstrap')


def test_parse_arguments_config_with_no_subaction_errors():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('config')


def test_parse_arguments_config_with_help_shows_config_help(capsys):
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit) as exit:
        module.parse_arguments('config', '--help')

    assert exit.value.code == 0
    captured = capsys.readouterr()
    assert 'global arguments:' not in captured.out
    assert 'config arguments:' in captured.out
    assert 'config sub-actions:' in captured.out


def test_parse_arguments_config_with_subaction_but_missing_flags_errors():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit) as exit:
        module.parse_arguments('config', 'bootstrap')

    assert exit.value.code == 2


def test_parse_arguments_config_with_subaction_and_help_shows_subaction_help(capsys):
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(SystemExit) as exit:
        module.parse_arguments('config', 'bootstrap', '--help')

    assert exit.value.code == 0
    captured = capsys.readouterr()
    assert 'config bootstrap arguments:' in captured.out


def test_parse_arguments_config_with_subaction_and_required_flags_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('config', 'bootstrap', '--repository', 'repo.borg')


def test_parse_arguments_config_with_subaction_and_global_flags_at_start_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('--verbosity', '1', 'config', 'bootstrap', '--repository', 'repo.borg')


def test_parse_arguments_config_with_subaction_and_global_flags_at_end_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('config', 'bootstrap', '--repository', 'repo.borg', '--verbosity', '1')


def test_parse_arguments_config_with_subaction_and_explicit_config_file_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments(
        'config', 'bootstrap', '--repository', 'repo.borg', '--config', 'test.yaml'
    )


def test_parse_arguments_with_borg_action_and_dry_run_raises():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    with pytest.raises(ValueError):
        module.parse_arguments('--dry-run', 'borg', 'list')


def test_parse_arguments_with_borg_action_and_no_dry_run_does_not_raise():
    flexmock(module.collect).should_receive('get_default_config_paths').and_return(['default'])

    module.parse_arguments('borg', 'list')



import os
import sys
from io import StringIO

import pytest
from flexmock import flexmock

from borgmatic.config import validate as module


def test_schema_filename_finds_schema_path():
    schema_path = '/var/borgmatic/config/schema.yaml'

    flexmock(os.path).should_receive('dirname').and_return('/var/borgmatic/config')
    builtins = flexmock(sys.modules['builtins'])
    builtins.should_receive('open').with_args(schema_path).and_return(StringIO())
    assert module.schema_filename() == schema_path


def test_schema_filename_raises_filenotfounderror():
    schema_path = '/var/borgmatic/config/schema.yaml'

    flexmock(os.path).should_receive('dirname').and_return('/var/borgmatic/config')
    builtins = flexmock(sys.modules['builtins'])
    builtins.should_receive('open').with_args(schema_path).and_raise(FileNotFoundError)

    with pytest.raises(FileNotFoundError):
        module.schema_filename()


def test_format_json_error_path_element_formats_array_index():
    module.format_json_error_path_element(3) == '[3]'


def test_format_json_error_path_element_formats_property():
    module.format_json_error_path_element('foo') == '.foo'


def test_format_json_error_formats_error_including_path():
    flexmock(module).format_json_error_path_element = lambda element: f'.{element}'
    error = flexmock(message='oops', path=['foo', 'bar'])

    assert module.format_json_error(error) == "At 'foo.bar': oops"


def test_format_json_error_formats_error_without_path():
    flexmock(module).should_receive('format_json_error_path_element').never()
    error = flexmock(message='oops', path=[])

    assert module.format_json_error(error) == 'At the top level: oops'


def test_validation_error_string_contains_errors():
    flexmock(module).format_json_error = lambda error: error.message
    error = module.Validation_error('config.yaml', ('oops', 'uh oh'))

    result = str(error)

    assert 'config.yaml' in result
    assert 'oops' in result
    assert 'uh oh' in result


def test_apply_logical_validation_raises_if_unknown_repository_in_check_repositories():
    flexmock(module).format_json_error = lambda error: error.message

    with pytest.raises(module.Validation_error):
        module.apply_logical_validation(
            'config.yaml',
            {
                'repositories': ['repo.borg', 'other.borg'],
                'keep_secondly': 1000,
                'check_repositories': ['repo.borg', 'unknown.borg'],
            },
        )


def test_apply_logical_validation_does_not_raise_if_known_repository_path_in_check_repositories():
    module.apply_logical_validation(
        'config.yaml',
        {
            'repositories': [{'path': 'repo.borg'}, {'path': 'other.borg'}],
            'keep_secondly': 1000,
            'check_repositories': ['repo.borg'],
        },
    )


def test_apply_logical_validation_does_not_raise_if_known_repository_label_in_check_repositories():
    module.apply_logical_validation(
        'config.yaml',
        {
            'repositories': [
                {'path': 'repo.borg', 'label': 'my_repo'},
                {'path': 'other.borg', 'label': 'other_repo'},
            ],
            'keep_secondly': 1000,
            'check_repositories': ['my_repo'],
        },
    )


def test_apply_logical_validation_does_not_raise_if_archive_name_format_and_prefix_present():
    module.apply_logical_validation(
        'config.yaml',
        {
            'archive_name_format': '{hostname}-{now}',  # noqa: FS003
            'prefix': '{hostname}-',  # noqa: FS003
            'prefix': '{hostname}-',  # noqa: FS003
        },
    )


def test_apply_logical_validation_does_not_raise_otherwise():
    module.apply_logical_validation('config.yaml', {'keep_secondly': 1000})


def test_normalize_repository_path_passes_through_remote_repository():
    repository = 'example.org:test.borg'

    module.normalize_repository_path(repository) == repository


def test_normalize_repository_path_passes_through_file_repository():
    repository = 'file:///foo/bar/test.borg'
    flexmock(module.os.path).should_receive('abspath').and_return('/foo/bar/test.borg')

    module.normalize_repository_path(repository) == '/foo/bar/test.borg'


def test_normalize_repository_path_passes_through_absolute_repository():
    repository = '/foo/bar/test.borg'
    flexmock(module.os.path).should_receive('abspath').and_return(repository)

    module.normalize_repository_path(repository) == repository


def test_normalize_repository_path_resolves_relative_repository():
    repository = 'test.borg'
    absolute = '/foo/bar/test.borg'
    flexmock(module.os.path).should_receive('abspath').and_return(absolute)

    module.normalize_repository_path(repository) == absolute


def test_repositories_match_does_not_raise():
    flexmock(module).should_receive('normalize_repository_path')

    module.repositories_match('foo', 'bar')


def test_guard_configuration_contains_repository_does_not_raise_when_repository_in_config():
    flexmock(module).should_receive('repositories_match').replace_with(
        lambda first, second: first == second
    )

    module.guard_configuration_contains_repository(
        repository='repo', configurations={'config.yaml': {'repositories': ['repo']}}
    )


def test_guard_configuration_contains_repository_does_not_raise_when_repository_label_in_config():
    module.guard_configuration_contains_repository(
        repository='repo',
        configurations={'config.yaml': {'repositories': [{'path': 'foo/bar', 'label': 'repo'}]}},
    )


def test_guard_configuration_contains_repository_does_not_raise_when_repository_not_given():
    module.guard_configuration_contains_repository(
        repository=None, configurations={'config.yaml': {'repositories': ['repo']}}
    )


def test_guard_configuration_contains_repository_errors_when_repository_missing_from_config():
    flexmock(module).should_receive('repositories_match').replace_with(
        lambda first, second: first == second
    )

    with pytest.raises(ValueError):
        module.guard_configuration_contains_repository(
            repository='nope',
            configurations={'config.yaml': {'repositories': ['repo', 'repo2']}},
        )


def test_guard_single_repository_selected_raises_when_multiple_repositories_configured_and_none_selected():
    with pytest.raises(ValueError):
        module.guard_single_repository_selected(
            repository=None,
            configurations={'config.yaml': {'repositories': ['repo', 'repo2']}},
        )


def test_guard_single_repository_selected_does_not_raise_when_single_repository_configured_and_none_selected():
    module.guard_single_repository_selected(
        repository=None,
        configurations={'config.yaml': {'repositories': ['repo']}},
    )


def test_guard_single_repository_selected_does_not_raise_when_no_repositories_configured_and_one_selected():
    module.guard_single_repository_selected(
        repository='repo',
        configurations={'config.yaml': {'repositories': []}},
    )


def test_guard_single_repository_selected_does_not_raise_when_repositories_configured_and_one_selected():
    module.guard_single_repository_selected(
        repository='repo',
        configurations={'config.yaml': {'repositories': ['repo', 'repo2']}},
    )



import logging

from flexmock import flexmock

from borgmatic.borg import rinfo as module

from ..test_verbosity import insert_logging_mock


def test_display_repository_info_calls_borg_with_flags():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--json', '--repo', 'repo'),
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'rinfo', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )

    module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
    )


def test_display_repository_info_without_borg_features_calls_borg_with_info_sub_command():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--json', 'repo'),
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'info', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )

    module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
    )


def test_display_repository_info_with_log_info_calls_borg_with_info_flag():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--info', '--json', '--repo', 'repo'),
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'rinfo', '--info', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )
    insert_logging_mock(logging.INFO)
    module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
    )


def test_display_repository_info_with_log_info_and_json_suppresses_most_borg_output():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags').never()

    insert_logging_mock(logging.INFO)
    json_output = module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=True),
        global_arguments=flexmock(log_json=False),
    )

    assert json_output == '[]'


def test_display_repository_info_with_log_debug_calls_borg_with_debug_flag():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--debug', '--show-rc', '--json', '--repo', 'repo'),
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'rinfo', '--debug', '--show-rc', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )
    insert_logging_mock(logging.DEBUG)

    module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
    )


def test_display_repository_info_with_log_debug_and_json_suppresses_most_borg_output():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags').never()

    insert_logging_mock(logging.DEBUG)
    json_output = module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=True),
        global_arguments=flexmock(log_json=False),
    )

    assert json_output == '[]'


def test_display_repository_info_with_json_calls_borg_with_json_flag():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags').never()

    json_output = module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=True),
        global_arguments=flexmock(log_json=False),
    )

    assert json_output == '[]'


def test_display_repository_info_with_local_path_calls_borg_via_local_path():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg1', 'rinfo', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg1', 'rinfo', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg1',
        borg_exit_codes=None,
        extra_environment=None,
    )

    module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
        local_path='borg1',
    )


def test_display_repository_info_with_exit_codes_calls_borg_using_them():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    borg_exit_codes = flexmock()
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=borg_exit_codes,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'rinfo', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=borg_exit_codes,
        extra_environment=None,
    )

    module.display_repository_info(
        repository_path='repo',
        config={'borg_exit_codes': borg_exit_codes},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
    )


def test_display_repository_info_with_remote_path_calls_borg_with_remote_path_flags():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--remote-path', 'borg1', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'rinfo', '--remote-path', 'borg1', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )

    module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
        remote_path='borg1',
    )


def test_display_repository_info_with_log_json_calls_borg_with_log_json_flags():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--log-json', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'rinfo', '--log-json', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )

    module.display_repository_info(
        repository_path='repo',
        config={},
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=True),
    )


def test_display_repository_info_with_lock_wait_calls_borg_with_lock_wait_flags():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module.logging).ANSWER = module.borgmatic.logger.ANSWER
    config = {'lock_wait': 5}
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'rinfo', '--lock-wait', '5', '--json', '--repo', 'repo'),
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('[]')
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'rinfo', '--lock-wait', '5', '--repo', 'repo'),
        output_log_level=module.borgmatic.logger.ANSWER,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )

    module.display_repository_info(
        repository_path='repo',
        config=config,
        local_borg_version='2.3.4',
        rinfo_arguments=flexmock(json=False),
        global_arguments=flexmock(log_json=False),
    )



import logging
import subprocess
import time

import pytest
from flexmock import flexmock

import borgmatic.hooks.command
from borgmatic.commands import borgmatic as module


@pytest.mark.parametrize(
    'config,arguments,expected_actions',
    (
        ({}, {}, []),
        ({'skip_actions': []}, {}, []),
        ({'skip_actions': ['prune', 'check']}, {}, ['prune', 'check']),
        (
            {'skip_actions': ['prune', 'check']},
            {'check': flexmock(force=False)},
            ['prune', 'check'],
        ),
        ({'skip_actions': ['prune', 'check']}, {'check': flexmock(force=True)}, ['prune']),
    ),
)
def test_get_skip_actions_uses_config_and_arguments(config, arguments, expected_actions):
    assert module.get_skip_actions(config, arguments) == expected_actions


def test_run_configuration_runs_actions_for_each_repository():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    expected_results = [flexmock(), flexmock()]
    flexmock(module).should_receive('run_actions').and_return(expected_results[:1]).and_return(
        expected_results[1:]
    )
    config = {'repositories': [{'path': 'foo'}, {'path': 'bar'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1)}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == expected_results


def test_run_configuration_with_skip_actions_does_not_raise():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return(['compact'])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module).should_receive('run_actions').and_return(flexmock()).and_return(flexmock())
    config = {'repositories': [{'path': 'foo'}, {'path': 'bar'}], 'skip_actions': ['compact']}
    arguments = {'global': flexmock(monitoring_verbosity=1)}

    list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))


def test_run_configuration_with_invalid_borg_version_errors():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_raise(ValueError)
    flexmock(module.command).should_receive('execute_hook').never()
    flexmock(module.dispatch).should_receive('call_hooks').never()
    flexmock(module).should_receive('run_actions').never()
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'prune': flexmock()}

    list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))


def test_run_configuration_logs_monitor_start_error():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.dispatch).should_receive('call_hooks').and_raise(OSError).and_return(
        None
    ).and_return(None).and_return(None)
    expected_results = [flexmock()]
    flexmock(module).should_receive('log_error_records').and_return(expected_results)
    flexmock(module).should_receive('run_actions').never()
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == expected_results


def test_run_configuration_bails_for_monitor_start_soft_failure():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    error = subprocess.CalledProcessError(borgmatic.hooks.command.SOFT_FAIL_EXIT_CODE, 'try again')
    flexmock(module.dispatch).should_receive('call_hooks').and_raise(error)
    flexmock(module).should_receive('log_error_records').never()
    flexmock(module).should_receive('run_actions').never()
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == []


def test_run_configuration_logs_actions_error():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module.dispatch).should_receive('call_hooks')
    expected_results = [flexmock()]
    flexmock(module).should_receive('log_error_records').and_return(expected_results)
    flexmock(module).should_receive('run_actions').and_raise(OSError)
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False)}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == expected_results


def test_run_configuration_skips_remaining_actions_for_actions_soft_failure_but_still_pings_monitor():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.dispatch).should_receive('call_hooks').times(5)
    error = subprocess.CalledProcessError(borgmatic.hooks.command.SOFT_FAIL_EXIT_CODE, 'try again')
    flexmock(module).should_receive('run_actions').and_raise(error)
    flexmock(module).should_receive('log_error_records').never()
    flexmock(module.command).should_receive('considered_soft_failure').and_return(True)
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == []


def test_run_configuration_logs_monitor_log_error():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.dispatch).should_receive('call_hooks').and_return(None).and_return(
        None
    ).and_raise(OSError)
    expected_results = [flexmock()]
    flexmock(module).should_receive('log_error_records').and_return(expected_results)
    flexmock(module).should_receive('run_actions').and_return([])
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == expected_results


def test_run_configuration_still_pings_monitor_for_monitor_log_soft_failure():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    error = subprocess.CalledProcessError(borgmatic.hooks.command.SOFT_FAIL_EXIT_CODE, 'try again')
    flexmock(module.dispatch).should_receive('call_hooks').and_return(None).and_return(
        None
    ).and_raise(error).and_return(None).and_return(None).times(5)
    flexmock(module).should_receive('log_error_records').never()
    flexmock(module).should_receive('run_actions').and_return([])
    flexmock(module.command).should_receive('considered_soft_failure').and_return(True)
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == []


def test_run_configuration_logs_monitor_finish_error():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.dispatch).should_receive('call_hooks').and_return(None).and_return(
        None
    ).and_return(None).and_raise(OSError)
    expected_results = [flexmock()]
    flexmock(module).should_receive('log_error_records').and_return(expected_results)
    flexmock(module).should_receive('run_actions').and_return([])
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == expected_results


def test_run_configuration_bails_for_monitor_finish_soft_failure():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    error = subprocess.CalledProcessError(borgmatic.hooks.command.SOFT_FAIL_EXIT_CODE, 'try again')
    flexmock(module.dispatch).should_receive('call_hooks').and_return(None).and_return(
        None
    ).and_raise(None).and_raise(error)
    flexmock(module).should_receive('log_error_records').never()
    flexmock(module).should_receive('run_actions').and_return([])
    flexmock(module.command).should_receive('considered_soft_failure').and_return(True)
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == []


def test_run_configuration_does_not_call_monitoring_hooks_if_monitoring_hooks_are_disabled():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(module.DISABLED)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())

    flexmock(module.dispatch).should_receive('call_hooks').never()
    flexmock(module).should_receive('run_actions').and_return([])

    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=-2, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == []


def test_run_configuration_logs_on_error_hook_error():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook').and_raise(OSError)
    expected_results = [flexmock(), flexmock()]
    flexmock(module).should_receive('log_error_records').and_return(
        expected_results[:1]
    ).and_return(expected_results[1:])
    flexmock(module).should_receive('run_actions').and_raise(OSError)
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == expected_results


def test_run_configuration_bails_for_on_error_hook_soft_failure():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    error = subprocess.CalledProcessError(borgmatic.hooks.command.SOFT_FAIL_EXIT_CODE, 'try again')
    flexmock(module.command).should_receive('execute_hook').and_raise(error)
    expected_results = [flexmock()]
    flexmock(module).should_receive('log_error_records').and_return(expected_results)
    flexmock(module).should_receive('run_actions').and_raise(OSError)
    config = {'repositories': [{'path': 'foo'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}

    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))

    assert results == expected_results


def test_run_configuration_retries_soft_error():
    # Run action first fails, second passes.
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module).should_receive('run_actions').and_raise(OSError).and_return([])
    flexmock(module).should_receive('log_error_records').and_return([flexmock()]).once()
    config = {'repositories': [{'path': 'foo'}], 'retries': 1}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == []


def test_run_configuration_retries_hard_error():
    # Run action fails twice.
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module).should_receive('run_actions').and_raise(OSError).times(2)
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()])
    error_logs = [flexmock()]
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
    ).and_return(error_logs)
    config = {'repositories': [{'path': 'foo'}], 'retries': 1}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == error_logs


def test_run_configuration_repos_ordered():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module).should_receive('run_actions').and_raise(OSError).times(2)
    expected_results = [flexmock(), flexmock()]
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository', OSError
    ).and_return(expected_results[:1]).ordered()
    flexmock(module).should_receive('log_error_records').with_args(
        'bar: Error running actions for repository', OSError
    ).and_return(expected_results[1:]).ordered()
    config = {'repositories': [{'path': 'foo'}, {'path': 'bar'}]}
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == expected_results


def test_run_configuration_retries_round_robin():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module).should_receive('run_actions').and_raise(OSError).times(4)
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()
    flexmock(module).should_receive('log_error_records').with_args(
        'bar: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()
    foo_error_logs = [flexmock()]
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository', OSError
    ).and_return(foo_error_logs).ordered()
    bar_error_logs = [flexmock()]
    flexmock(module).should_receive('log_error_records').with_args(
        'bar: Error running actions for repository', OSError
    ).and_return(bar_error_logs).ordered()
    config = {
        'repositories': [{'path': 'foo'}, {'path': 'bar'}],
        'retries': 1,
    }
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == foo_error_logs + bar_error_logs


def test_run_configuration_retries_one_passes():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module).should_receive('run_actions').and_raise(OSError).and_raise(OSError).and_return(
        []
    ).and_raise(OSError).times(4)
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()
    flexmock(module).should_receive('log_error_records').with_args(
        'bar: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return(flexmock()).ordered()
    error_logs = [flexmock()]
    flexmock(module).should_receive('log_error_records').with_args(
        'bar: Error running actions for repository', OSError
    ).and_return(error_logs).ordered()
    config = {
        'repositories': [{'path': 'foo'}, {'path': 'bar'}],
        'retries': 1,
    }
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == error_logs


def test_run_configuration_retry_wait():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module).should_receive('run_actions').and_raise(OSError).times(4)
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()

    flexmock(time).should_receive('sleep').with_args(10).and_return().ordered()
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()

    flexmock(time).should_receive('sleep').with_args(20).and_return().ordered()
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()

    flexmock(time).should_receive('sleep').with_args(30).and_return().ordered()
    error_logs = [flexmock()]
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository', OSError
    ).and_return(error_logs).ordered()
    config = {
        'repositories': [{'path': 'foo'}],
        'retries': 3,
        'retry_wait': 10,
    }
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == error_logs


def test_run_configuration_retries_timeout_multiple_repos():
    flexmock(module).should_receive('verbosity_to_log_level').and_return(logging.INFO)
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module).should_receive('run_actions').and_raise(OSError).and_raise(OSError).and_return(
        []
    ).and_raise(OSError).times(4)
    flexmock(module).should_receive('log_error_records').with_args(
        'foo: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()
    flexmock(module).should_receive('log_error_records').with_args(
        'bar: Error running actions for repository',
        OSError,
        levelno=logging.WARNING,
        log_command_error_output=True,
    ).and_return([flexmock()]).ordered()

    # Sleep before retrying foo (and passing)
    flexmock(time).should_receive('sleep').with_args(10).and_return().ordered()

    # Sleep before retrying bar (and failing)
    flexmock(time).should_receive('sleep').with_args(10).and_return().ordered()
    error_logs = [flexmock()]
    flexmock(module).should_receive('log_error_records').with_args(
        'bar: Error running actions for repository', OSError
    ).and_return(error_logs).ordered()
    config = {
        'repositories': [{'path': 'foo'}, {'path': 'bar'}],
        'retries': 1,
        'retry_wait': 10,
    }
    arguments = {'global': flexmock(monitoring_verbosity=1, dry_run=False), 'create': flexmock()}
    results = list(module.run_configuration('test.yaml', config, ['/tmp/test.yaml'], arguments))
    assert results == error_logs


def test_run_actions_runs_rcreate():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.rcreate).should_receive('run_rcreate').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'rcreate': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_adds_log_file_to_hook_context():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    expected = flexmock()
    flexmock(borgmatic.actions.create).should_receive('run_create').with_args(
        config_filename=object,
        repository={'path': 'repo'},
        config={'repositories': []},
        config_paths=[],
        hook_context={'repository': 'repo', 'repositories': '', 'log_file': 'foo'},
        local_borg_version=object,
        create_arguments=object,
        global_arguments=object,
        dry_run_label='',
        local_path=object,
        remote_path=object,
    ).once().and_return(expected)

    result = tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'create': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )
    assert result == (expected,)


def test_run_actions_runs_transfer():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.transfer).should_receive('run_transfer').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'transfer': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_create():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    expected = flexmock()
    flexmock(borgmatic.actions.create).should_receive('run_create').and_yield(expected).once()

    result = tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'create': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )
    assert result == (expected,)


def test_run_actions_with_skip_actions_skips_create():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return(['create'])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.create).should_receive('run_create').never()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'create': flexmock()},
            config_filename=flexmock(),
            config={'repositories': [], 'skip_actions': ['create']},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_prune():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.prune).should_receive('run_prune').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'prune': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_with_skip_actions_skips_prune():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return(['prune'])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.prune).should_receive('run_prune').never()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'prune': flexmock()},
            config_filename=flexmock(),
            config={'repositories': [], 'skip_actions': ['prune']},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_compact():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.compact).should_receive('run_compact').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'compact': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_with_skip_actions_skips_compact():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return(['compact'])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.compact).should_receive('run_compact').never()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'compact': flexmock()},
            config_filename=flexmock(),
            config={'repositories': [], 'skip_actions': ['compact']},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_check_when_repository_enabled_for_checks():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module.checks).should_receive('repository_enabled_for_checks').and_return(True)
    flexmock(borgmatic.actions.check).should_receive('run_check').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'check': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_skips_check_when_repository_not_enabled_for_checks():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module.checks).should_receive('repository_enabled_for_checks').and_return(False)
    flexmock(borgmatic.actions.check).should_receive('run_check').never()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'check': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_with_skip_actions_skips_check():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return(['check'])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(module.checks).should_receive('repository_enabled_for_checks').and_return(True)
    flexmock(borgmatic.actions.check).should_receive('run_check').never()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'check': flexmock()},
            config_filename=flexmock(),
            config={'repositories': [], 'skip_actions': ['check']},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_extract():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.extract).should_receive('run_extract').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'extract': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_export_tar():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.export_tar).should_receive('run_export_tar').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'export-tar': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_mount():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.mount).should_receive('run_mount').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'mount': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_restore():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.restore).should_receive('run_restore').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'restore': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_rlist():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    expected = flexmock()
    flexmock(borgmatic.actions.rlist).should_receive('run_rlist').and_yield(expected).once()

    result = tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'rlist': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )
    assert result == (expected,)


def test_run_actions_runs_list():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    expected = flexmock()
    flexmock(borgmatic.actions.list).should_receive('run_list').and_yield(expected).once()

    result = tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'list': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )
    assert result == (expected,)


def test_run_actions_runs_rinfo():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    expected = flexmock()
    flexmock(borgmatic.actions.rinfo).should_receive('run_rinfo').and_yield(expected).once()

    result = tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'rinfo': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )
    assert result == (expected,)


def test_run_actions_runs_info():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    expected = flexmock()
    flexmock(borgmatic.actions.info).should_receive('run_info').and_yield(expected).once()

    result = tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'info': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )
    assert result == (expected,)


def test_run_actions_runs_break_lock():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.break_lock).should_receive('run_break_lock').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'break-lock': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_export_key():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.export_key).should_receive('run_export_key').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'export': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_borg():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.borg).should_receive('run_borg').once()

    tuple(
        module.run_actions(
            arguments={'global': flexmock(dry_run=False, log_file='foo'), 'borg': flexmock()},
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_run_actions_runs_multiple_actions_in_argument_order():
    flexmock(module).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('get_skip_actions').and_return([])
    flexmock(module.command).should_receive('execute_hook')
    flexmock(borgmatic.actions.borg).should_receive('run_borg').once().ordered()
    flexmock(borgmatic.actions.restore).should_receive('run_restore').once().ordered()

    tuple(
        module.run_actions(
            arguments={
                'global': flexmock(dry_run=False, log_file='foo'),
                'borg': flexmock(),
                'restore': flexmock(),
            },
            config_filename=flexmock(),
            config={'repositories': []},
            config_paths=[],
            local_path=flexmock(),
            remote_path=flexmock(),
            local_borg_version=flexmock(),
            repository={'path': 'repo'},
        )
    )


def test_load_configurations_collects_parsed_configurations_and_logs():
    configuration = flexmock()
    other_configuration = flexmock()
    test_expected_logs = [flexmock(), flexmock()]
    other_expected_logs = [flexmock(), flexmock()]
    flexmock(module.validate).should_receive('parse_configuration').and_return(
        configuration, ['/tmp/test.yaml'], test_expected_logs
    ).and_return(other_configuration, ['/tmp/other.yaml'], other_expected_logs)

    configs, config_paths, logs = tuple(module.load_configurations(('test.yaml', 'other.yaml')))

    assert configs == {'test.yaml': configuration, 'other.yaml': other_configuration}
    assert config_paths == ['/tmp/other.yaml', '/tmp/test.yaml']
    assert set(logs) >= set(test_expected_logs + other_expected_logs)


def test_load_configurations_logs_warning_for_permission_error():
    flexmock(module.validate).should_receive('parse_configuration').and_raise(PermissionError)

    configs, config_paths, logs = tuple(module.load_configurations(('test.yaml',)))

    assert configs == {}
    assert config_paths == []
    assert max(log.levelno for log in logs) == logging.WARNING


def test_load_configurations_logs_critical_for_parse_error():
    flexmock(module.validate).should_receive('parse_configuration').and_raise(ValueError)

    configs, config_paths, logs = tuple(module.load_configurations(('test.yaml',)))

    assert configs == {}
    assert config_paths == []
    assert max(log.levelno for log in logs) == logging.CRITICAL


def test_log_record_does_not_raise():
    module.log_record(levelno=1, foo='bar', baz='quux')


def test_log_record_with_suppress_does_not_raise():
    module.log_record(levelno=1, foo='bar', baz='quux', suppress_log=True)


def test_log_error_records_generates_output_logs_for_message_only():
    flexmock(module).should_receive('log_record').replace_with(dict).once()

    logs = tuple(module.log_error_records('Error'))

    assert {log['levelno'] for log in logs} == {logging.CRITICAL}


def test_log_error_records_generates_output_logs_for_called_process_error_with_bytes_ouput():
    flexmock(module).should_receive('log_record').replace_with(dict).times(3)
    flexmock(module.logger).should_receive('getEffectiveLevel').and_return(logging.WARNING)

    logs = tuple(
        module.log_error_records('Error', subprocess.CalledProcessError(1, 'ls', b'error output'))
    )

    assert {log['levelno'] for log in logs} == {logging.CRITICAL}
    assert any(log for log in logs if 'error output' in str(log))


def test_log_error_records_generates_output_logs_for_called_process_error_with_string_ouput():
    flexmock(module).should_receive('log_record').replace_with(dict).times(3)
    flexmock(module.logger).should_receive('getEffectiveLevel').and_return(logging.WARNING)

    logs = tuple(
        module.log_error_records('Error', subprocess.CalledProcessError(1, 'ls', 'error output'))
    )

    assert {log['levelno'] for log in logs} == {logging.CRITICAL}
    assert any(log for log in logs if 'error output' in str(log))


def test_log_error_records_splits_called_process_error_with_multiline_ouput_into_multiple_logs():
    flexmock(module).should_receive('log_record').replace_with(dict).times(4)
    flexmock(module.logger).should_receive('getEffectiveLevel').and_return(logging.WARNING)

    logs = tuple(
        module.log_error_records(
            'Error', subprocess.CalledProcessError(1, 'ls', 'error output\nanother line')
        )
    )

    assert {log['levelno'] for log in logs} == {logging.CRITICAL}
    assert any(log for log in logs if 'error output' in str(log))


def test_log_error_records_generates_logs_for_value_error():
    flexmock(module).should_receive('log_record').replace_with(dict).twice()

    logs = tuple(module.log_error_records('Error', ValueError()))

    assert {log['levelno'] for log in logs} == {logging.CRITICAL}


def test_log_error_records_generates_logs_for_os_error():
    flexmock(module).should_receive('log_record').replace_with(dict).twice()

    logs = tuple(module.log_error_records('Error', OSError()))

    assert {log['levelno'] for log in logs} == {logging.CRITICAL}


def test_log_error_records_generates_nothing_for_other_error():
    flexmock(module).should_receive('log_record').never()

    logs = tuple(module.log_error_records('Error', KeyError()))

    assert logs == ()


def test_get_local_path_uses_configuration_value():
    assert module.get_local_path({'test.yaml': {'local_path': 'borg1'}}) == 'borg1'


def test_get_local_path_without_local_path_defaults_to_borg():
    assert module.get_local_path({'test.yaml': {}}) == 'borg'


def test_collect_highlander_action_summary_logs_info_for_success_with_bootstrap():
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.borgmatic.actions.config.bootstrap).should_receive('run_bootstrap')
    arguments = {
        'bootstrap': flexmock(repository='repo'),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=False
        )
    )
    assert {log.levelno for log in logs} == {logging.ANSWER}


def test_collect_highlander_action_summary_logs_error_on_bootstrap_failure():
    flexmock(module.borg_version).should_receive('local_borg_version').and_return(flexmock())
    flexmock(module.borgmatic.actions.config.bootstrap).should_receive('run_bootstrap').and_raise(
        ValueError
    )
    arguments = {
        'bootstrap': flexmock(repository='repo'),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=False
        )
    )

    assert {log.levelno for log in logs} == {logging.CRITICAL}


def test_collect_highlander_action_summary_logs_error_on_bootstrap_local_borg_version_failure():
    flexmock(module.borg_version).should_receive('local_borg_version').and_raise(ValueError)
    flexmock(module.borgmatic.actions.config.bootstrap).should_receive('run_bootstrap').never()
    arguments = {
        'bootstrap': flexmock(repository='repo'),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=False
        )
    )

    assert {log.levelno for log in logs} == {logging.CRITICAL}


def test_collect_highlander_action_summary_logs_info_for_success_with_generate():
    flexmock(module.borgmatic.actions.config.generate).should_receive('run_generate')
    arguments = {
        'generate': flexmock(destination='test.yaml'),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=False
        )
    )
    assert {log.levelno for log in logs} == {logging.ANSWER}


def test_collect_highlander_action_summary_logs_error_on_generate_failure():
    flexmock(module.borgmatic.actions.config.generate).should_receive('run_generate').and_raise(
        ValueError
    )
    arguments = {
        'generate': flexmock(destination='test.yaml'),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=False
        )
    )

    assert {log.levelno for log in logs} == {logging.CRITICAL}


def test_collect_highlander_action_summary_logs_info_for_success_with_validate():
    flexmock(module.borgmatic.actions.config.validate).should_receive('run_validate')
    arguments = {
        'validate': flexmock(),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=False
        )
    )
    assert {log.levelno for log in logs} == {logging.ANSWER}


def test_collect_highlander_action_summary_logs_error_on_validate_parse_failure():
    flexmock(module.borgmatic.actions.config.validate).should_receive('run_validate')
    arguments = {
        'validate': flexmock(),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=True
        )
    )

    assert {log.levelno for log in logs} == {logging.CRITICAL}


def test_collect_highlander_action_summary_logs_error_on_run_validate_failure():
    flexmock(module.borgmatic.actions.config.validate).should_receive('run_validate').and_raise(
        ValueError
    )
    arguments = {
        'validate': flexmock(),
        'global': flexmock(dry_run=False),
    }

    logs = tuple(
        module.collect_highlander_action_summary_logs(
            {'test.yaml': {}}, arguments=arguments, configuration_parse_errors=False
        )
    )

    assert {log.levelno for log in logs} == {logging.CRITICAL}


def test_collect_configuration_run_summary_logs_info_for_success():
    flexmock(module.command).should_receive('execute_hook').never()
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return([])
    arguments = {}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert {log.levelno for log in logs} == {logging.INFO}


def test_collect_configuration_run_summary_executes_hooks_for_create():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return([])
    arguments = {'create': flexmock(), 'global': flexmock(monitoring_verbosity=1, dry_run=False)}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert {log.levelno for log in logs} == {logging.INFO}


def test_collect_configuration_run_summary_logs_info_for_success_with_extract():
    flexmock(module.validate).should_receive('guard_single_repository_selected')
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return([])
    arguments = {'extract': flexmock(repository='repo')}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert {log.levelno for log in logs} == {logging.INFO}


def test_collect_configuration_run_summary_logs_extract_with_repository_error():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository').and_raise(
        ValueError
    )
    expected_logs = (flexmock(),)
    flexmock(module).should_receive('log_error_records').and_return(expected_logs)
    arguments = {'extract': flexmock(repository='repo')}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert logs == expected_logs


def test_collect_configuration_run_summary_logs_info_for_success_with_mount():
    flexmock(module.validate).should_receive('guard_single_repository_selected')
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return([])
    arguments = {'mount': flexmock(repository='repo')}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert {log.levelno for log in logs} == {logging.INFO}


def test_collect_configuration_run_summary_logs_mount_with_repository_error():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository').and_raise(
        ValueError
    )
    expected_logs = (flexmock(),)
    flexmock(module).should_receive('log_error_records').and_return(expected_logs)
    arguments = {'mount': flexmock(repository='repo')}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert logs == expected_logs


def test_collect_configuration_run_summary_logs_missing_configs_error():
    arguments = {'global': flexmock(config_paths=[])}
    expected_logs = (flexmock(),)
    flexmock(module).should_receive('log_error_records').and_return(expected_logs)

    logs = tuple(
        module.collect_configuration_run_summary_logs({}, config_paths=[], arguments=arguments)
    )

    assert logs == expected_logs


def test_collect_configuration_run_summary_logs_pre_hook_error():
    flexmock(module.command).should_receive('execute_hook').and_raise(ValueError)
    expected_logs = (flexmock(),)
    flexmock(module).should_receive('log_error_records').and_return(expected_logs)
    arguments = {'create': flexmock(), 'global': flexmock(monitoring_verbosity=1, dry_run=False)}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert logs == expected_logs


def test_collect_configuration_run_summary_logs_post_hook_error():
    flexmock(module.command).should_receive('execute_hook').and_return(None).and_raise(ValueError)
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return([])
    expected_logs = (flexmock(),)
    flexmock(module).should_receive('log_error_records').and_return(expected_logs)
    arguments = {'create': flexmock(), 'global': flexmock(monitoring_verbosity=1, dry_run=False)}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert expected_logs[0] in logs


def test_collect_configuration_run_summary_logs_for_list_with_archive_and_repository_error():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository').and_raise(
        ValueError
    )
    expected_logs = (flexmock(),)
    flexmock(module).should_receive('log_error_records').and_return(expected_logs)
    arguments = {'list': flexmock(repository='repo', archive='test')}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert logs == expected_logs


def test_collect_configuration_run_summary_logs_info_for_success_with_list():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return([])
    arguments = {'list': flexmock(repository='repo', archive=None)}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert {log.levelno for log in logs} == {logging.INFO}


def test_collect_configuration_run_summary_logs_run_configuration_error():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return(
        [logging.makeLogRecord(dict(levelno=logging.CRITICAL, levelname='CRITICAL', msg='Error'))]
    )
    flexmock(module).should_receive('log_error_records').and_return([])
    arguments = {}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert {log.levelno for log in logs} == {logging.CRITICAL}


def test_collect_configuration_run_summary_logs_run_umount_error():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return([])
    flexmock(module.borg_umount).should_receive('unmount_archive').and_raise(OSError)
    flexmock(module).should_receive('log_error_records').and_return(
        [logging.makeLogRecord(dict(levelno=logging.CRITICAL, levelname='CRITICAL', msg='Error'))]
    )
    arguments = {'umount': flexmock(mount_point='/mnt')}

    logs = tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}}, config_paths=['/tmp/test.yaml'], arguments=arguments
        )
    )

    assert {log.levelno for log in logs} == {logging.INFO, logging.CRITICAL}


def test_collect_configuration_run_summary_logs_outputs_merged_json_results():
    flexmock(module.validate).should_receive('guard_configuration_contains_repository')
    flexmock(module).should_receive('run_configuration').and_return(['foo', 'bar']).and_return(
        ['baz']
    )
    stdout = flexmock()
    stdout.should_receive('write').with_args('["foo", "bar", "baz"]').once()
    flexmock(module.sys).stdout = stdout
    arguments = {}

    tuple(
        module.collect_configuration_run_summary_logs(
            {'test.yaml': {}, 'test2.yaml': {}},
            config_paths=['/tmp/test.yaml', '/tmp/test2.yaml'],
            arguments=arguments,
        )
    )



import pkgutil

import borgmatic.actions
import borgmatic.config.load
import borgmatic.config.validate

MAXIMUM_LINE_LENGTH = 80


def test_schema_line_length_stays_under_limit():
    schema_file = open('borgmatic/config/schema.yaml')

    for line in schema_file.readlines():
        assert len(line.rstrip('\n')) <= MAXIMUM_LINE_LENGTH


ACTIONS_MODULE_NAMES_TO_OMIT = {'arguments', 'export_key', 'json'}
ACTIONS_MODULE_NAMES_TO_ADD = {'key', 'umount'}


def test_schema_skip_actions_correspond_to_supported_actions():
    '''
    Ensure that the allowed actions in the schema's "skip_actions" option don't drift from
    borgmatic's actual supported actions.
    '''
    schema = borgmatic.config.load.load_configuration(borgmatic.config.validate.schema_filename())
    schema_skip_actions = set(schema['properties']['skip_actions']['items']['enum'])
    supported_actions = {
        module.name.replace('_', '-')
        for module in pkgutil.iter_modules(borgmatic.actions.__path__)
        if module.name not in ACTIONS_MODULE_NAMES_TO_OMIT
    }.union(ACTIONS_MODULE_NAMES_TO_ADD)

    assert schema_skip_actions == supported_actions



import logging

from flexmock import flexmock

from borgmatic.borg import mount as module

from ..test_verbosity import insert_logging_mock


def insert_execute_command_mock(command, borg_exit_codes=None):
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command').with_args(
        command,
        borg_local_path=command[0],
        borg_exit_codes=borg_exit_codes,
        extra_environment=None,
    ).once()


def test_mount_archive_calls_borg_with_required_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'mount', 'repo', '/mnt'))

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive=None,
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_with_borg_features_calls_borg_with_repository_and_match_archives_flags():
    flexmock(module.feature).should_receive('available').and_return(True)
    flexmock(module.flags).should_receive('make_repository_flags').and_return(
        (
            '--repo',
            'repo',
        )
    )
    insert_execute_command_mock(
        ('borg', 'mount', '--repo', 'repo', '--match-archives', 'archive', '/mnt')
    )

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_without_archive_calls_borg_with_repository_flags_only():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', 'repo::archive', '/mnt'))

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_calls_borg_with_path_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', 'repo::archive', '/mnt', 'path1', 'path2'))

    mount_arguments = flexmock(
        mount_point='/mnt', options=None, paths=['path1', 'path2'], foreground=False
    )
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_calls_borg_with_local_path():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg1', 'mount', 'repo::archive', '/mnt'))

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
        local_path='borg1',
    )


def test_mount_archive_calls_borg_using_exit_codes():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    borg_exit_codes = flexmock()
    insert_execute_command_mock(
        ('borg', 'mount', 'repo::archive', '/mnt'),
        borg_exit_codes=borg_exit_codes,
    )

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={'borg_exit_codes': borg_exit_codes},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_calls_borg_with_remote_path_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(
        ('borg', 'mount', '--remote-path', 'borg1', 'repo::archive', '/mnt')
    )

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
        remote_path='borg1',
    )


def test_mount_archive_calls_borg_with_umask_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', '--umask', '0770', 'repo::archive', '/mnt'))

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={'umask': '0770'},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_calls_borg_with_log_json_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', '--log-json', 'repo::archive', '/mnt'))

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=True),
    )


def test_mount_archive_calls_borg_with_lock_wait_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', '--lock-wait', '5', 'repo::archive', '/mnt'))

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={'lock_wait': '5'},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_with_log_info_calls_borg_with_info_parameter():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', '--info', 'repo::archive', '/mnt'))
    insert_logging_mock(logging.INFO)

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_with_log_debug_calls_borg_with_debug_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', '--debug', '--show-rc', 'repo::archive', '/mnt'))
    insert_logging_mock(logging.DEBUG)

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=False)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_calls_borg_with_foreground_parameter():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command').with_args(
        ('borg', 'mount', '--foreground', 'repo::archive', '/mnt'),
        output_file=module.DO_NOT_CAPTURE,
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    ).once()

    mount_arguments = flexmock(mount_point='/mnt', options=None, paths=None, foreground=True)
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_calls_borg_with_options_flags():
    flexmock(module.feature).should_receive('available').and_return(False)
    flexmock(module.flags).should_receive('make_repository_archive_flags').and_return(
        ('repo::archive',)
    )
    insert_execute_command_mock(('borg', 'mount', '-o', 'super_mount', 'repo::archive', '/mnt'))

    mount_arguments = flexmock(
        mount_point='/mnt', options='super_mount', paths=None, foreground=False
    )
    module.mount_archive(
        repository_path='repo',
        archive='archive',
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_mount_archive_with_date_based_matching_calls_borg_with_date_based_flags():
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(
        (
            '--newer',
            '1d',
            '--newest',
            '1y',
            '--older',
            '1m',
            '--oldest',
            '1w',
            '--match-archives',
            None,
        )
    )
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('--repo', 'repo'))
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command').with_args(
        (
            'borg',
            'mount',
            '--newer',
            '1d',
            '--newest',
            '1y',
            '--older',
            '1m',
            '--oldest',
            '1w',
            '--match-archives',
            None,
            '--repo',
            'repo',
            '/mnt',
        ),
        borg_local_path='borg',
        borg_exit_codes=None,
        extra_environment=None,
    )

    mount_arguments = flexmock(
        mount_point='/mnt',
        options=None,
        paths=None,
        foreground=False,
        newer='1d',
        newest='1y',
        older='1m',
        oldest='1w',
    )
    module.mount_archive(
        repository_path='repo',
        archive=None,
        mount_arguments=mount_arguments,
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )



from flexmock import flexmock

from borgmatic.actions import compact as module


def test_compact_actions_calls_hooks_for_configured_repository():
    flexmock(module.logger).answer = lambda message: None
    flexmock(module.borgmatic.borg.feature).should_receive('available').and_return(True)
    flexmock(module.borgmatic.config.validate).should_receive('repositories_match').never()
    flexmock(module.borgmatic.borg.compact).should_receive('compact_segments').once()
    flexmock(module.borgmatic.hooks.command).should_receive('execute_hook').times(2)
    compact_arguments = flexmock(
        repository=None, progress=flexmock(), cleanup_commits=flexmock(), threshold=flexmock()
    )
    global_arguments = flexmock(monitoring_verbosity=1, dry_run=False)

    module.run_compact(
        config_filename='test.yaml',
        repository={'path': 'repo'},
        config={},
        hook_context={},
        local_borg_version=None,
        compact_arguments=compact_arguments,
        global_arguments=global_arguments,
        dry_run_label='',
        local_path=None,
        remote_path=None,
    )


def test_compact_runs_with_selected_repository():
    flexmock(module.logger).answer = lambda message: None
    flexmock(module.borgmatic.config.validate).should_receive(
        'repositories_match'
    ).once().and_return(True)
    flexmock(module.borgmatic.borg.feature).should_receive('available').and_return(True)
    flexmock(module.borgmatic.borg.compact).should_receive('compact_segments').once()
    compact_arguments = flexmock(
        repository=flexmock(), progress=flexmock(), cleanup_commits=flexmock(), threshold=flexmock()
    )
    global_arguments = flexmock(monitoring_verbosity=1, dry_run=False)

    module.run_compact(
        config_filename='test.yaml',
        repository={'path': 'repo'},
        config={},
        hook_context={},
        local_borg_version=None,
        compact_arguments=compact_arguments,
        global_arguments=global_arguments,
        dry_run_label='',
        local_path=None,
        remote_path=None,
    )


def test_compact_bails_if_repository_does_not_match():
    flexmock(module.logger).answer = lambda message: None
    flexmock(module.borgmatic.borg.feature).should_receive('available').and_return(True)
    flexmock(module.borgmatic.config.validate).should_receive(
        'repositories_match'
    ).once().and_return(False)
    flexmock(module.borgmatic.borg.compact).should_receive('compact_segments').never()
    compact_arguments = flexmock(
        repository=flexmock(), progress=flexmock(), cleanup_commits=flexmock(), threshold=flexmock()
    )
    global_arguments = flexmock(monitoring_verbosity=1, dry_run=False)

    module.run_compact(
        config_filename='test.yaml',
        repository={'path': 'repo'},
        config={},
        hook_context={},
        local_borg_version=None,
        compact_arguments=compact_arguments,
        global_arguments=global_arguments,
        dry_run_label='',
        local_path=None,
        remote_path=None,
    )



from flexmock import flexmock

from borgmatic.actions import rlist as module


def test_run_rlist_does_not_raise():
    flexmock(module.logger).answer = lambda message: None
    flexmock(module.borgmatic.config.validate).should_receive('repositories_match').and_return(True)
    flexmock(module.borgmatic.borg.rlist).should_receive('list_repository')
    rlist_arguments = flexmock(repository=flexmock(), json=False)

    list(
        module.run_rlist(
            repository={'path': 'repo'},
            config={},
            local_borg_version=None,
            rlist_arguments=rlist_arguments,
            global_arguments=flexmock(),
            local_path=None,
            remote_path=None,
        )
    )


def test_run_rlist_produces_json():
    flexmock(module.logger).answer = lambda message: None
    flexmock(module.borgmatic.config.validate).should_receive('repositories_match').and_return(True)
    flexmock(module.borgmatic.borg.rlist).should_receive('list_repository').and_return(flexmock())
    parsed_json = flexmock()
    flexmock(module.borgmatic.actions.json).should_receive('parse_json').and_return(parsed_json)
    rlist_arguments = flexmock(repository=flexmock(), json=True)

    assert list(
        module.run_rlist(
            repository={'path': 'repo'},
            config={},
            local_borg_version=None,
            rlist_arguments=rlist_arguments,
            global_arguments=flexmock(),
            local_path=None,
            remote_path=None,
        )
    ) == [parsed_json]



import argparse
import logging

import pytest
from flexmock import flexmock

from borgmatic.borg import rlist as module

from ..test_verbosity import insert_logging_mock

BORG_LIST_LATEST_ARGUMENTS = (
    '--last',
    '1',
    '--short',
    'repo',
)


def test_resolve_archive_name_passes_through_non_latest_archive_name():
    archive = 'myhost-2030-01-01T14:41:17.647620'

    assert (
        module.resolve_archive_name(
            'repo',
            archive,
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
        )
        == archive
    )


def test_resolve_archive_name_calls_borg_with_flags():
    expected_archive = 'archive-name'
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return(expected_archive + '\n')

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
        )
        == expected_archive
    )


def test_resolve_archive_name_with_log_info_calls_borg_without_info_flag():
    expected_archive = 'archive-name'
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return(expected_archive + '\n')
    insert_logging_mock(logging.INFO)

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
        )
        == expected_archive
    )


def test_resolve_archive_name_with_log_debug_calls_borg_without_debug_flag():
    expected_archive = 'archive-name'
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return(expected_archive + '\n')
    insert_logging_mock(logging.DEBUG)

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
        )
        == expected_archive
    )


def test_resolve_archive_name_with_local_path_calls_borg_via_local_path():
    expected_archive = 'archive-name'
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg1', 'list') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg1',
        borg_exit_codes=None,
    ).and_return(expected_archive + '\n')

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
            local_path='borg1',
        )
        == expected_archive
    )


def test_resolve_archive_name_with_exit_codes_calls_borg_using_them():
    expected_archive = 'archive-name'
    flexmock(module.environment).should_receive('make_environment')
    borg_exit_codes = flexmock()
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=borg_exit_codes,
    ).and_return(expected_archive + '\n')

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={'borg_exit_codes': borg_exit_codes},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
        )
        == expected_archive
    )


def test_resolve_archive_name_with_remote_path_calls_borg_with_remote_path_flags():
    expected_archive = 'archive-name'
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list', '--remote-path', 'borg1') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return(expected_archive + '\n')

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
            remote_path='borg1',
        )
        == expected_archive
    )


def test_resolve_archive_name_without_archives_raises():
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return('')

    with pytest.raises(ValueError):
        module.resolve_archive_name(
            'repo',
            'latest',
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
        )


def test_resolve_archive_name_with_log_json_calls_borg_with_log_json_flags():
    expected_archive = 'archive-name'

    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list', '--log-json') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return(expected_archive + '\n')

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=True),
        )
        == expected_archive
    )


def test_resolve_archive_name_with_lock_wait_calls_borg_with_lock_wait_flags():
    expected_archive = 'archive-name'

    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').with_args(
        ('borg', 'list', '--lock-wait', 'okay') + BORG_LIST_LATEST_ARGUMENTS,
        extra_environment=None,
        borg_local_path='borg',
        borg_exit_codes=None,
    ).and_return(expected_archive + '\n')

    assert (
        module.resolve_archive_name(
            'repo',
            'latest',
            config={'lock_wait': 'okay'},
            local_borg_version='1.2.3',
            global_arguments=flexmock(log_json=False),
        )
        == expected_archive
    )


def test_make_rlist_command_includes_log_info():
    insert_logging_mock(logging.INFO)
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--info', 'repo')


def test_make_rlist_command_includes_json_but_not_info():
    insert_logging_mock(logging.INFO)
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(('--json',))
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=True, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--json', 'repo')


def test_make_rlist_command_includes_log_debug():
    insert_logging_mock(logging.DEBUG)
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--debug', '--show-rc', 'repo')


def test_make_rlist_command_includes_json_but_not_debug():
    insert_logging_mock(logging.DEBUG)
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(('--json',))
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=True, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--json', 'repo')


def test_make_rlist_command_includes_json():
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(('--json',))
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=True, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--json', 'repo')


def test_make_rlist_command_includes_log_json():
    flexmock(module.flags).should_receive('make_flags').and_return(()).and_return(
        ('--log-json',)
    ).and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=True),
    )

    assert command == ('borg', 'list', '--log-json', 'repo')


def test_make_rlist_command_includes_lock_wait():
    flexmock(module.flags).should_receive('make_flags').and_return(()).and_return(
        ('--lock-wait', '5')
    ).and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={'lock_wait': 5},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--lock-wait', '5', 'repo')


def test_make_rlist_command_includes_local_path():
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
        local_path='borg2',
    )

    assert command == ('borg2', 'list', 'repo')


def test_make_rlist_command_includes_remote_path():
    flexmock(module.flags).should_receive('make_flags').and_return(
        ('--remote-path', 'borg2')
    ).and_return(()).and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
        remote_path='borg2',
    )

    assert command == ('borg', 'list', '--remote-path', 'borg2', 'repo')


def test_make_rlist_command_transforms_prefix_into_match_archives():
    flexmock(module.flags).should_receive('make_flags').and_return(()).and_return(()).and_return(
        ('--match-archives', 'sh:foo*')
    )
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(archive=None, paths=None, json=False, prefix='foo'),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--match-archives', 'sh:foo*', 'repo')


def test_make_rlist_command_prefers_prefix_over_archive_name_format():
    flexmock(module.flags).should_receive('make_flags').and_return(()).and_return(()).and_return(
        ('--match-archives', 'sh:foo*')
    )
    flexmock(module.flags).should_receive('make_match_archives_flags').never()
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={'archive_name_format': 'bar-{now}'},  # noqa: FS003
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(archive=None, paths=None, json=False, prefix='foo'),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--match-archives', 'sh:foo*', 'repo')


def test_make_rlist_command_transforms_archive_name_format_into_match_archives():
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, 'bar-{now}', '1.2.3'  # noqa: FS003
    ).and_return(('--match-archives', 'sh:bar-*'))
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={'archive_name_format': 'bar-{now}'},  # noqa: FS003
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--match-archives', 'sh:bar-*', 'repo')


def test_make_rlist_command_includes_short():
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(('--short',))
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None, paths=None, json=False, prefix=None, match_archives=None, short=True
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--short', 'repo')


@pytest.mark.parametrize(
    'argument_name',
    (
        'sort_by',
        'first',
        'last',
        'exclude',
        'exclude_from',
        'pattern',
        'patterns_from',
    ),
)
def test_make_rlist_command_includes_additional_flags(argument_name):
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(
        (f"--{argument_name.replace('_', '-')}", 'value')
    )
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None,
            paths=None,
            json=False,
            prefix=None,
            match_archives=None,
            find_paths=None,
            format=None,
            **{argument_name: 'value'},
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--' + argument_name.replace('_', '-'), 'value', 'repo')


def test_make_rlist_command_with_match_archives_calls_borg_with_match_archives_flags():
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        'foo-*',
        None,
        '1.2.3',
    ).and_return(('--match-archives', 'foo-*'))
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(())
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None,
            paths=None,
            json=False,
            prefix=None,
            match_archives='foo-*',
            find_paths=None,
            format=None,
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == ('borg', 'list', '--match-archives', 'foo-*', 'repo')


def test_list_repository_calls_two_commands():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('make_rlist_command')
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command_and_capture_output').once()
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags')
    flexmock(module).should_receive('execute_command').once()

    module.list_repository(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=argparse.Namespace(json=False),
        global_arguments=flexmock(),
    )


def test_list_repository_with_json_calls_json_command_only():
    flexmock(module.borgmatic.logger).should_receive('add_custom_log_levels')
    flexmock(module).should_receive('make_rlist_command')
    flexmock(module.environment).should_receive('make_environment')
    json_output = flexmock()
    flexmock(module).should_receive('execute_command_and_capture_output').and_return(json_output)
    flexmock(module.flags).should_receive('warn_for_aggressive_archive_flags').never()
    flexmock(module).should_receive('execute_command').never()

    assert (
        module.list_repository(
            repository_path='repo',
            config={},
            local_borg_version='1.2.3',
            rlist_arguments=argparse.Namespace(json=True),
            global_arguments=flexmock(),
        )
        == json_output
    )


def test_make_rlist_command_with_date_based_matching_calls_borg_with_date_based_flags():
    flexmock(module.flags).should_receive('make_flags').and_return(())
    flexmock(module.flags).should_receive('make_match_archives_flags').with_args(
        None, None, '1.2.3'
    ).and_return(())
    flexmock(module.flags).should_receive('make_flags_from_arguments').and_return(
        ('--newer', '1d', '--newest', '1y', '--older', '1m', '--oldest', '1w')
    )
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))

    command = module.make_rlist_command(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        rlist_arguments=flexmock(
            archive=None,
            paths=None,
            json=False,
            prefix=None,
            match_archives=None,
            newer='1d',
            newest='1y',
            older='1m',
            oldest='1w',
        ),
        global_arguments=flexmock(log_json=False),
    )

    assert command == (
        'borg',
        'list',
        '--newer',
        '1d',
        '--newest',
        '1y',
        '--older',
        '1m',
        '--oldest',
        '1w',
        'repo',
    )



from borgmatic.borg import feature as module


def test_available_true_for_new_enough_borg_version():
    assert module.available(module.Feature.COMPACT, '1.3.7')


def test_available_true_for_borg_version_introducing_feature():
    assert module.available(module.Feature.COMPACT, '1.2.0a2')


def test_available_true_for_borg_stable_version_introducing_feature():
    assert module.available(module.Feature.COMPACT, '1.2.0')


def test_available_false_for_too_old_borg_version():
    assert not module.available(module.Feature.COMPACT, '1.1.5')



import logging

from flexmock import flexmock

from borgmatic.borg import break_lock as module

from ..test_verbosity import insert_logging_mock


def insert_execute_command_mock(command, borg_exit_codes=None):
    flexmock(module.environment).should_receive('make_environment')
    flexmock(module).should_receive('execute_command').with_args(
        command,
        borg_local_path=command[0],
        borg_exit_codes=borg_exit_codes,
        extra_environment=None,
    ).once()


def test_break_lock_calls_borg_with_required_flags():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'break-lock', 'repo'))

    module.break_lock(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_break_lock_calls_borg_with_local_path():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg1', 'break-lock', 'repo'))

    module.break_lock(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
        local_path='borg1',
    )


def test_break_lock_calls_borg_using_exit_codes():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg1', 'break-lock', 'repo'))

    module.break_lock(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
        local_path='borg1',
    )


def test_break_lock_calls_borg_with_remote_path_flags():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'break-lock', '--remote-path', 'borg1', 'repo'))

    module.break_lock(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
        remote_path='borg1',
    )


def test_break_lock_calls_borg_with_umask_flags():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'break-lock', '--umask', '0770', 'repo'))

    module.break_lock(
        repository_path='repo',
        config={'umask': '0770'},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_break_lock_calls_borg_with_log_json_flags():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'break-lock', '--log-json', 'repo'))

    module.break_lock(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=True),
    )


def test_break_lock_calls_borg_with_lock_wait_flags():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'break-lock', '--lock-wait', '5', 'repo'))

    module.break_lock(
        repository_path='repo',
        config={'lock_wait': '5'},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_break_lock_with_log_info_calls_borg_with_info_parameter():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'break-lock', '--info', 'repo'))
    insert_logging_mock(logging.INFO)

    module.break_lock(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )


def test_break_lock_with_log_debug_calls_borg_with_debug_flags():
    flexmock(module.flags).should_receive('make_repository_flags').and_return(('repo',))
    insert_execute_command_mock(('borg', 'break-lock', '--debug', '--show-rc', 'repo'))
    insert_logging_mock(logging.DEBUG)

    module.break_lock(
        repository_path='repo',
        config={},
        local_borg_version='1.2.3',
        global_arguments=flexmock(log_json=False),
    )
